#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "record.h"

struct event {
	unsigned long tm;
	float x, y, z;
	float theta, phi;

	struct event *next;
};

static struct event *head, *tail;
static struct event *iter;
static int num_events;

void rec_reset(void)
{
	while(head) {
		iter = head;
		head = head->next;
		free(iter);
	}
	num_events = 0;
}

int rec_load(const char *fname)
{
	FILE *fp;
	int i;

	if(!(fp = fopen(fname, "rb"))) {
		fprintf(stderr, "failed to open recording: %s: %s\n", fname, strerror(errno));
		return -1;
	}
	fread(&num_events, sizeof num_events, 1, fp);

	head = tail = 0;

	for(i=0; i<num_events; i++) {
		struct event *ev = malloc(sizeof *ev);
		if(!ev) {
			perror("failed to allocate event structure");
			fclose(fp);
			return -1;
		}
		if(fread(ev, sizeof *ev, 1, fp) < 1) {
			perror("unexpected end of file");
			fclose(fp);
			return -1;
		}

		if(!head) {
			head = tail = ev;
		} else {
			tail->next = ev;
			tail = ev;
		}
		ev->next = 0;
	}
	iter = 0;

	fclose(fp);
	return 0;
}

int rec_save(const char *fname)
{
	FILE *fp;
	struct event *ev;

	if(!(fp = fopen(fname, "wb"))) {
		fprintf(stderr, "failed to open recording: %s: %s\n", fname, strerror(errno));
		return -1;
	}
	fwrite(&num_events, sizeof num_events, 1, fp);

	ev = head;
	while(ev) {
		fwrite(ev, sizeof *ev, 1, fp);
		ev = ev->next;
	}

	fclose(fp);
	return 0;
}

void rec_add(unsigned long tm, float x, float y, float z, float theta, float phi)
{
	struct event *ev = malloc(sizeof *ev);
	if(!ev) {
		fprintf(stderr, "failed to append rec-event: %s\n", strerror(errno));
		return;
	}
	ev->tm = tm;
	ev->x = x;
	ev->y = y;
	ev->z = z;
	ev->theta = theta;
	ev->phi = phi;

	ev->next = 0;
	if(head) {
		tail->next = ev;
		tail = ev;
	} else {
		head = tail = ev;
	}

	num_events++;
}

void rec_get(unsigned long tm, float *x, float *y, float *z, float *theta, float *phi)
{
	if(!iter || iter->tm > tm) {
		iter = head;
	}

	while(iter && iter->next) {
		if(iter->next->tm > tm) {
			break;
		}
		iter = iter->next;
	}

	if(iter) {
		*x = iter->x;
		*y = iter->y;
		*z = iter->z;
		*theta = iter->theta;
		*phi = iter->phi;
	}
}
